/*
 *
 * docint.c
 * Description:
 * DOCSIS driver interrupt implementation
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#include <asm/uaccess.h>	/* copy_to/from_user */
#include <linux/init.h>		/* modules */
#include <linux/module.h>	/* module */
#include <linux/device.h>	/* module */
#include <linux/types.h>
#include <linux/fs.h>		/* chrdev allocation */
#include <linux/slab.h>		/* kmalloc and kfree */
#include <linux/cdev.h>
#include <linux/wait.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/errno.h>
#include "hal_interrupt_cause.h"  
#include "docint.h"
#include "pal.h"

#define MAXIIT	1024
#define	FIRST_DOC_IRQ			MAC_ERR_INT_IRQ_NUM
#define	MIN2IRQ(m)				((MINOR((m)->di_cdev.dev) == 5)? PHY_PGA_GRT_INT_NUM : (MINOR((m)->di_cdev.dev) + FIRST_DOC_IRQ) )

MODULE_LICENSE("GPL");

#define PUMA5_HW_REV_1_X   0
#define PUMA5_HW_REV_2_0   1

#define MAC_KERNEL_BASE_ADDRESS                     0xD9000000

/* Global variable that defines the IRQ number for the PGA GRT interrupt */
unsigned int pgaGrtIntNum;


void docint_read_register(Uint32  addr, Uint32 *value)
{
   volatile Uint32 *reg;

   reg = (Uint32 *) addr;	
   PAL_sysCacheInvalidate(PAL_OSMEM_ADDR_DAT, (void*)(reg), sizeof(Uint32));
   
   *value = *(reg);
}

static void docint_print_dsid_stats(void)
{
    unsigned int  regVal1, regVal2, regVal3, regVal4;
    unsigned char dsid;

    printk("DSID\tOUT-OF-RANGE\tMISSED\tFORWARDED   WARN-THRESHOLD\n");
    printk("----\t------------\t------\t---------   --------------\n");
    for (dsid = 0; dsid < 16; dsid++)
    {
        docint_read_register(MAC_KERNEL_BASE_ADDRESS + 0x4BE00 + (0x10 * dsid), &regVal1);
        docint_read_register(MAC_KERNEL_BASE_ADDRESS + 0x4BE04 + (0x10 * dsid), &regVal2);
        docint_read_register(MAC_KERNEL_BASE_ADDRESS + 0x4BE08 + (0x10 * dsid), &regVal3);
        docint_read_register(MAC_KERNEL_BASE_ADDRESS + 0x4BE0C + (0x10 * dsid), &regVal4);
        printk("%d\t%d\t\t%d\t%d\t    %d\n", dsid, regVal1, regVal2, regVal3, regVal4);
    }
    printk("\n");
}


/* count and pos are ignored since an ii_t is always read */
static ssize_t
docint_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
	di_t *dev = filp->private_data;	/* dev was stored in filp during open */

	if (down_interruptible(&dev->di_sem))
		return -ERESTARTSYS;
	while (dev->di_roffset == dev->di_woffset) {
		up(&dev->di_sem);
		if (wait_event_interruptible(dev->di_rqueue, dev->di_roffset != dev->di_woffset))
			return -ERESTARTSYS;
		if (down_interruptible(&dev->di_sem))
			return -ERESTARTSYS;
	}
	if (copy_to_user(buf, dev->di_data + dev->di_roffset, sizeof(ii_t))) {
		up(&dev->di_sem);
		return -EFAULT;
	}
	dev->di_roffset++;
	dev->di_roffset %= MAXIIT;
	atomic_inc(&dev->di_free);
	up(&dev->di_sem);
	*f_pos += sizeof(ii_t);
	return sizeof(ii_t);
}

static ssize_t
docint_write(struct file *filp, const char __user *buf, size_t count, loff_t *f_pos)
{
	di_t *dev = filp->private_data;
	ii_t info;
	ssize_t	retval;

	if (down_interruptible(&dev->di_sem))
		return -ERESTARTSYS;
	if (copy_from_user(&info, buf, sizeof(ii_t)))
		retval = -EFAULT;
	else {
		*f_pos += sizeof(ii_t);
		retval = sizeof(ii_t);
		HAL_IsrMaskInterrupt(info.ii_inum, info.ii_masked);   
	}
	up(&dev->di_sem);

	return retval;
}

static unsigned int
docint_poll(struct file *filp, struct poll_table_struct *wait)
{
	di_t *dev = filp->private_data;
	unsigned int mask = 0;

	/*
	 * The buffer is circular. It is considered full if
	 * di_woffset is right behind di_roffset and empty
	 * if the two are equal.
	 */
	down(&dev->di_sem);
	poll_wait(filp, &dev->di_rqueue, wait);
	if (dev->di_roffset != dev->di_woffset) {
		printk(KERN_INFO "docint_poll irq %d waked up\n", MIN2IRQ(dev));
		mask |= POLLIN | POLLRDNORM;	/* readable */
	}
	up(&dev->di_sem);

	return mask;
}

/* interrupt cause table */
extern dih_t docitbl[];

int
docint_cbregister(unsigned cause, char masked, void(*callback)(unsigned))
{
	if (cause >= MAX_CAUSE) {
		printk(KERN_WARNING "docint_cbregister: bad cause\n");
		return 0;
	}
	docitbl[cause].dih_func = callback;
	docitbl[cause].dih_masked = masked;

	return 1;
}


/*
 * This routine discovers what's the real interrupt cause
 * It also clears the interrupt
 */
static inline int
get_cause(int irq, IntCauseErrorGroup_t *buf, int *group)
{
    /* this can not be part of the "switch/case" below */
    /* because PHY_PGA_GRT_INT_NUM is not constant.    */
    if(irq == PHY_PGA_GRT_INT_NUM)
    {
		*(unsigned int *)buf = PHY_PGA_GRT_INT_MASK;
		*group = PHY_PGA_GRT_INT_GROUP;
		return 32;
    }

	switch (irq) {
	case MAC_ERR_INT_IRQ_NUM:
		HAL_IsrGetCauseMacErrorGroup((IntCauseErrorGroup_t *)buf);
		*group = MAC_ERR1_INT_GROUP;
		return 128;

	case MAC_UCD_INT_IRQ_NUM:
		*(unsigned int *)buf = HAL_IsrGetCauseUcdChange();
		*group = MAC_UCD_INT_GROUP;
		break;

	case MAC_US_FW_INT_IRQ_NUM:
		*(unsigned int *)buf = HAL_IsrGetCauseUsFw();
		*group = MAC_US_FW_INT_GROUP;
		break;

	case MAC_DS_FW_INT_IRQ_NUM:
		*(unsigned int *)buf = HAL_IsrGetCauseDsFw();
		*group = MAC_DS_FW_INT_GROUP;
		break;

	case MAC_PHY_INT_IRQ_NUM:
        /* Currently, phy interrupts are handled by using bit 30 */
        /* of MAC US FW interrupts.                              */
        /**(unsigned int *)buf = HAL_IsrGetCausePhy();
        *group = PHY_INT_GROUP;*/
		break;

	default:
		printk(KERN_WARNING "get_cause default %d\n", irq);
		return 0;
	}

	return 32;
}


/*
 * The handler discovers what's the interrupt cause
 * and calls the proper routine. Such a routine may
 * activate a tasklet, but this is decided by its writter
 */
static irqreturn_t
doc_ihandler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned int cause, i, bits, group;
	IntCauseErrorGroup_t mask;

	memset(&mask, 0, sizeof (mask));
	printk(KERN_INFO "doc_ihandler irq %d \n", irq);
	bits = get_cause(irq, &mask, &group);

	for (i = find_first_bit((void *)&mask, bits); i < bits; i = find_next_bit((void *)&mask, bits, i + 1)) {
		cause = HAL_GET_INTERRUPT_CAUSE(group, i);
		printk(KERN_INFO "doc_ihandler irq %d cause %d\n", irq, cause);
		if (cause >= 0 && docitbl[cause].dih_func)
			docitbl[cause].dih_func(cause);
	}

	/* Workaround for the 2 signals problem */
    if(irq >= 32)
        *(volatile unsigned int *)(0xe4000284) = (1<<(irq-32)); //TBD
    else
        *(volatile unsigned int *)(0xe4000280) = (1<<irq); //TBD

	return IRQ_HANDLED;
}

static int
docint_open(struct inode *inode, struct file *filp)
{
	di_t *dev;
    unsigned long flags;
	
	dev = container_of(inode->i_cdev, di_t, di_cdev);
	down(&dev->di_sem);
	printk(KERN_INFO "docint_open dev %p ref %d\n", dev, dev->ref);
	dev->ref++;
	filp->private_data = dev;	/* store for future use */

/*
	if (dev->di_irq != -1) {
		up(&dev->di_sem);
		return 0;
	}
*/
	if (dev->ref > 1) {
		up(&dev->di_sem);
		return 0;
	}
	dev->di_irq = MIN2IRQ(dev);

    /* for the PGA GRT interrupt - need specific initialization */
    if(dev->di_irq == PHY_PGA_GRT_INT_NUM)
        flags = IRQ_TYPE_EDGE_RISING;
    else 
        flags = 0;

	if (request_irq(dev->di_irq, doc_ihandler, flags, DI_NAME, (void *)dev))
		printk(KERN_WARNING "Can't get assigned irq %i\n", dev->di_irq);


	printk(KERN_INFO "request_irq %d in docint_open\n", dev->di_irq);
	up(&dev->di_sem);

	return 0;
}

static int
docint_release(struct inode *inode, struct file *filp)
{
	di_t *dev;

	dev = container_of(inode->i_cdev, di_t, di_cdev);
	down(&dev->di_sem);
	printk(KERN_INFO "docint_release: dev %p ref %d\n", dev, dev->ref);
	dev->ref--;
	if (dev->ref) {
		up(&dev->di_sem);
		return 0;
	}
	printk(KERN_INFO "docint_release: irq %d\n", dev->di_irq);
	free_irq(dev->di_irq, dev);
	dev->di_irq = -1;
	up(&dev->di_sem);
	return 0;
}

struct file_operations di_fops = {
	.owner		= THIS_MODULE,
	.read		= docint_read,
	.write		= docint_write,
	.poll		= docint_poll,
	.open		= docint_open,
	.release	= docint_release,
};

static dev_t dev; /* Contains major and first minor number */
static di_t *docint_devices;

static void
docint_cleanup(int count)
{
	int i;

	if (docint_devices) {
		for (i = 0; i < count; i++) {
			if (docint_devices[i].di_data)
				kfree(docint_devices[i].di_data);
			cdev_del(&docint_devices[i].di_cdev);
		}
		kfree(docint_devices);
	}
	unregister_chrdev_region(dev, DI_COUNT);
}

static void
initctab(di_t *dev)
{
	int i, j, group, irq = MIN2IRQ(dev);

	printk(KERN_INFO "initctab: irq %d dev 0x%x\n", irq, (Uint32)dev);


    /* this can not be part of the "switch/case" below */
    /* because PHY_PGA_GRT_INT_NUM is not constant.    */
    if(irq == PHY_PGA_GRT_INT_NUM)
    {
        /* This interrupt line only contains one interrupt (it is not really a group), */
        /* therefore the initialization method of the table is a little different.     */
		docitbl[HAL_GET_INTERRUPT_CAUSE(PHY_PGA_GRT_INT_GROUP, PHY_PGA_GRT_INT_INDEX)].dih_dev = dev;
		return; /* we do not want to continue in this function*/                                                                            
    }

	switch (irq) {
	case MAC_ERR_INT_IRQ_NUM:
		for (i = 0; i < MAC_ERROR_INT_BUNDLE_NUMBER; i++) {
			group = MAC_ERR1_INT_GROUP + i;
			for (j = 0; j < 32; j++)
				docitbl[HAL_GET_INTERRUPT_CAUSE(group, j)].dih_dev = dev;
		}
		return;

	case MAC_UCD_INT_IRQ_NUM:
		group = MAC_UCD_INT_GROUP;
		break;

	case MAC_US_FW_INT_IRQ_NUM:
		group = MAC_US_FW_INT_GROUP;
		break;

	case MAC_DS_FW_INT_IRQ_NUM:
		group = MAC_DS_FW_INT_GROUP;
		break;

	case MAC_PHY_INT_IRQ_NUM:
		group = PHY_INT_GROUP;
		break;

	default:
		printk(KERN_WARNING "initctab: bad irq %d\n", irq);
		return;
	}
	for (i = 0; i < 32; i++)
		docitbl[HAL_GET_INTERRUPT_CAUSE(group, i)].dih_dev = dev;
}

static struct class *class;

static int
docint_init(void)
{
	int i, r, devno;

	if ((r = alloc_chrdev_region(&dev, DI_MINOR_START, DI_COUNT, DI_NAME)) != 0) {
		printk(KERN_WARNING "docint_init: could not allocate device\n");
		return r;
	}

	class = class_create(THIS_MODULE, "DOCINT");

	printk(KERN_INFO "docint_init: registered with major number %i\n", MAJOR(dev));

	if ((docint_devices = kmalloc(DI_COUNT * sizeof(di_t), GFP_KERNEL)) == NULL) {
		printk(KERN_WARNING "docint_init: could not allocate %d devices\n", DI_COUNT);
		docint_cleanup(0);
		return -ENOMEM;
	}
	memset(docint_devices, 0, DI_COUNT * sizeof(di_t));

	for (i = 0; i < DI_COUNT; i++) {
		init_MUTEX(&docint_devices[i].di_sem);
		if ((docint_devices[i].di_data = kmalloc(MAXIIT * sizeof(ii_t), GFP_KERNEL)) == NULL) {
			printk(KERN_WARNING "docint_init: could not allocate data for device %d\n", i);
			docint_cleanup(i);
			return -ENOMEM;
		}
		docint_devices[i].di_roffset = 0;
		docint_devices[i].di_woffset = 0;
		docint_devices[i].ref = 0;
		atomic_set(&docint_devices[i].di_free, MAXIIT);
		init_waitqueue_head(&docint_devices[i].di_rqueue);

		cdev_init(&docint_devices[i].di_cdev, &di_fops);
		docint_devices[i].di_cdev.owner = THIS_MODULE;

		devno = MKDEV(MAJOR(dev), MINOR(dev) + i);
		if ((r = cdev_add(&docint_devices[i].di_cdev, devno, 1)) != 0)
			printk(KERN_WARNING "error %d adding %s%d\n", r, DI_NAME, i);
		else {
			printk(KERN_INFO "%s%d added\n", DI_NAME, i);
			initctab(&docint_devices[i]);
		}
		class_device_create(class, 0, MKDEV(MAJOR(dev), MINOR(dev) + i), 0, "DOCINT%d", i);
		docint_devices[i].di_irq = -1;
	}

    /* Resetting PHY-interrupts database */
    HAL_IsrResetPhyFecDB();

    /* Check the system revision and define the IRQ number for the PGA GRT interrupt */
    if (system_rev == PUMA5_HW_REV_2_0)
    {
        pgaGrtIntNum = PHY_PGA_GRT_INT_NUM_REV2;
    }   
    else
    {
        pgaGrtIntNum = PHY_PGA_GRT_INT_NUM_REV1;
        /* This is an external interrupt line that is muxed with GPIO #27. */
        /* Need to disable the GPIO in order to gain ext-interrupt functionality. */
        PAL_sysGpioCtrl(PHY_PGA_GRT_INT_REV1_GPIO, FUNCTIONAL_PIN, GPIO_INPUT_PIN);
    }


	return 0;
}

static void
docint_exit(void)
{
	docint_cleanup(DI_COUNT);
}


/****************************************************************************************************/
/*                                      KERNEL SIDE Interrupt Handlers                              */
/****************************************************************************************************/
void
docint_default_handler(unsigned int index)
{
#define DEBUG_PRINT_SIZE 20

	di_t *dev;
	ii_t info;
    unsigned int regVal;

    int i, intCount=0;

	info.ii_filler = 0;
	dev = docitbl[index].dih_dev;
	printk(KERN_INFO "Enter default_handler cause %d dev 0x%x\n", index, (Uint32)dev);
	/* Clear of the interrupt is done by get_cause() */

    if (!atomic_read(&dev->di_free))
    {
        HAL_IsrMaskInterrupt(208, True);

        /* Debug information for this scenario */
        printk("\ndev->di_irq = %d,  \n", dev->di_irq);
        printk("dev->di_roffset = %d,  \n", (int)dev->di_roffset);
        printk("dev->di_woffset = %d,  \n\n", (int)dev->di_woffset);

        printk("Last %d interrupt causes:\n", DEBUG_PRINT_SIZE);
        printk("=========================\n");

        for(i=(dev->di_woffset)-1; intCount < DEBUG_PRINT_SIZE ; i--, intCount++) 
        {
            if(i<0)
                i = MAXIIT-1;

            printk("    [%4d] cause = %d \n", i, dev->di_data[i].ii_inum);          
        }
        printk("\n");
        docint_print_dsid_stats();
        printk("\n");
        printk("Sleeping 1 second\n");
        msleep(1000);
        printk("\n");
        docint_print_dsid_stats();
        printk("\n\n\n");

        /* DSID and PDSP2 Registers */

        printk("DSID and PDSP2 Registers:\n");
        docint_read_register((MAC_KERNEL_BASE_ADDRESS+0x4a000), &regVal);
        printk("[0x%08X] 0x%08X\n", (MAC_KERNEL_BASE_ADDRESS+0x4a000), regVal);
        docint_read_register((MAC_KERNEL_BASE_ADDRESS+0x4bd80), &regVal);
        printk("[0x%08X] 0x%08X\n", (MAC_KERNEL_BASE_ADDRESS+0x4bd80), regVal);

        for (i = 0; (i < 0x7c); i+=4) {
            docint_read_register((MAC_KERNEL_BASE_ADDRESS+0x4a100+i), &regVal);
            if (!(i % 16))
                printk("\n0x%08X\t",MAC_KERNEL_BASE_ADDRESS+0x4a100+i);
            printk("0x%08X ", regVal);
        }
        printk("\n");
        docint_read_register((MAC_KERNEL_BASE_ADDRESS+0x4bd80), &regVal);
        printk("[0x%08X] 0x%08X\n", (MAC_KERNEL_BASE_ADDRESS+0x4bd80), regVal);

        for (i = 0; (i < 0x2fc); i+=4) {
            docint_read_register((MAC_KERNEL_BASE_ADDRESS+0x4b000+i), &regVal);
            if (!(i % 16))
                printk("\n0x%08X\t",MAC_KERNEL_BASE_ADDRESS+0x4b000+i);
            printk("0x%08X ", regVal);
        }
        printk("\n");
        docint_read_register((MAC_KERNEL_BASE_ADDRESS+0x4bd80), &regVal);
        printk("[0x%08X] 0x%08X\n", (MAC_KERNEL_BASE_ADDRESS+0x4bd80), regVal);

        for (i = 0; (i < 0xc); i+=4) {
            docint_read_register((MAC_KERNEL_BASE_ADDRESS+0x4bd00+i), &regVal);
            if (!(i % 16))
                printk("\n0x%08X\t",MAC_KERNEL_BASE_ADDRESS+0x4bd00+i);
            printk("0x%08X ", regVal);
        }
        printk("\n");
        docint_read_register((MAC_KERNEL_BASE_ADDRESS+0x4bd80), &regVal);
        printk("[0x%08X] 0x%08X\n", (MAC_KERNEL_BASE_ADDRESS+0x4bd80), regVal);

		panic("default_handler: no room for signal\n");
    }

	info.ii_inum = (short)index;
	info.ii_masked = docitbl[index].dih_masked;
	dev->di_data[dev->di_woffset++] = info;
	atomic_dec(&dev->di_free);
	dev->di_woffset %= MAXIIT;

	wake_up_interruptible(&dev->di_rqueue);
}

void
docint_mask_interrupt_handler(unsigned int index)
{
    HAL_IsrMaskInterrupt(index, True);

    docint_default_handler(index);
}


void
docint_phy_interrupts_handler (unsigned int index)
{
    unsigned int cause, i, bits, group;
    IntCauseErrorGroup_t mask;

    memset(&mask, 0, sizeof (mask));
    printk(KERN_INFO "--> docint_phy_interrupts_handler\n");
    group = PHY_INT_GROUP;
    bits  = 32;
    *(unsigned int *)&mask = HAL_IsrGetCausePhy();

    for (i = find_first_bit((void *)&mask, bits); i < bits; i = find_next_bit((void *)&mask, bits, i + 1)) {
        cause = HAL_GET_INTERRUPT_CAUSE(group, i);
        printk(KERN_INFO "--> docint_phy_interrupts_handler, cause %d\n", cause);
        if (cause >= 0 && docitbl[cause].dih_func)
            docitbl[cause].dih_func(cause);
    }
}



module_init(docint_init);
module_exit(docint_exit);

